package Network.Representation;

import Network.Representation.*;
import Network.Algorithm.*;

import java.util.*;

/**
 * A default implementation for the interface <code>Network</code>. Networks usually play the role of
 * a master over nodes and arcs. So the control of nodes and arcs are given to networks and users can add, delete
 * nodes and arcs via network structures.
 */
public class NetworkImplementation implements Network{
    
    int noOfNodes=0;
    int noOfArcs=0;
    int noOfAlgorithms=0;
    NodeList nodeList=null;
    NodeList nodeListEnd=null;
    //Here!
    static final NodeListFactory nodeListFactory = new NodeListFactory(200, 100);
    COM.objectspace.jgl.Array algorithmList;

    //Should create a NodeImplementationFactory that others can use
    //J.A.S., 8/31/98

    public NetworkImplementation(){
        algorithmList=new COM.objectspace.jgl.Array();
        algorithmList.trimToSize();
    }
    
    
    public final void clean(NodeFactory nodeFactory, ArcFactory arcFactory)
    {
        NodeList cleanList = nodeList;
        while (cleanList != null) {
            NodeList nextCleanList = cleanList.next;
            cleanList.elem.cleanInBoundArcs();
            cleanList.elem.cleanOutBoundArcs(arcFactory);
            cleanList.elem.cleanNodeAlgorithmInfo();
            if (nodeFactory != null) {
                cleanList.elem.setContainerNetwork(null);
                nodeFactory.collectNode(cleanList.elem);
            }
            else
                cleanList.elem = null; //Helps GC
            NetworkImplementation.nodeListFactory.unmake(cleanList);
            cleanList = nextCleanList;
        }
        nodeList = nodeListEnd = null;
        noOfNodes = 0;
        //Should clean algorithmList and set noOfAlgorithms = 0
    }
    
    public final int getNoOfNodes(){
        return noOfNodes;
    }

    public final int getNoOfArcs(){
        return noOfArcs;
    }
    
    /** 
     * Returns the number of algorithms that has their target_network as this network.
     * However since an algorithm may ave one helper algorithm (eg: simplex_algorithms have tree_algorithms in them)
     * this number may not reflect the number of algorithms that have been set by the user.
     */
    public final int getNoOfAlgorithms(){
        return noOfAlgorithms;
    }
    
    /**
     * Users should normally NEVER use this method
     */
    public void setNoOfAlgorithms(int setNo){
        noOfAlgorithms=setNo;
    }
    
    /**
     * Returns an enumeration of nodes in a network as <code>DList</code>.
     */
    public final NodeList getNodeList(){
        return nodeList;
    }
    
    public final NodeList getNodeListEnd(){
        return nodeListEnd;
    }
    
    /** 
     * Returns an enumeration of algorithms whose target_network is this network.
     */
    public final COM.objectspace.jgl.Array getAlgorithmList(){
        return algorithmList;
    }

    
    /**
     * Adds a node to this network. The node to be added to the network should NOT belong to any other network
     * or a <code>NodeError</code> will be thrown.
     *
     * @param addedNode the node that will be added to this network
     */
     
    public void addNode(Node addedNode) {
        NodeAlgInfo nodeInfo;
        int i;
        if ( addedNode.getContainerNetwork() != null ) {
            throw new NodeError("node can't be added: node already belongs to a network");
        }
        else {
            NodeList addedNodeList = NetworkImplementation.nodeListFactory.make(addedNode);
            if (nodeList != null)
                nodeListEnd.addAfter(addedNodeList);
            else
                nodeList = addedNodeList;
            nodeListEnd = addedNodeList;
            noOfNodes++;
            addedNode.setContainerNetwork(this);
            Vector nodeAlgInfo = addedNode.getNodeAlgorithmInfo();
            nodeAlgInfo.setSize(algorithmList.size());
            for (i=algorithmList.size(); i-- > 0;) {
                if (algorithmList.at(i)!=null) {
                    nodeInfo=((Algorithm)(algorithmList.at(i))).getInstanceOfNodeAlgInfo();
                    nodeAlgInfo.setElementAt(nodeInfo, i);
                }        
            }
        }
    }

    /**
     * Removes the node given by the parameter from this network. Failure to find the node in this network
     * results in NodeError to be thrown.
     *
     * @param deletedNode the node to be deleted from this network
     * @param nodeFactory the NodeFactory to which the node will be returned (can be null)
     * @param arcFactory  the ArcFactory to which the deletedNode's arcs will be returned (can be null)
     */
    public void removeNode(Node deletedNode, NodeFactory nodeFactory, ArcFactory arcFactory){
        
        NodeList nodeIterator=nodeList.find(deletedNode);
       
        ArcList arcIterator;
        
        if (nodeIterator != null) {
            arcIterator=deletedNode.getOutBoundArcs();
            while (arcIterator!=null) {
                removeArc(arcIterator.elem, arcFactory, false);
                arcIterator=arcIterator.next;
            }
            deletedNode.cleanOutBoundArcs(null); //Avoids double-collection with removeArc
            
            arcIterator=deletedNode.getInBoundArcs();
            while (arcIterator!=null) {
                removeArc(arcIterator.elem, arcFactory, true);
                arcIterator=arcIterator.next;
            }
            deletedNode.cleanInBoundArcs();
            
            deletedNode.setContainerNetwork(null);
            deletedNode.cleanNodeAlgorithmInfo();
            noOfNodes--;
            nodeIterator.remove();
            if (nodeList == nodeIterator)
                nodeList = nodeIterator.next;
            if (nodeListEnd == nodeIterator)
                nodeListEnd = nodeIterator.prev;
            if (nodeFactory != null)
                nodeFactory.collectNode(nodeIterator.elem);
            else
                nodeIterator.elem = null; //Helps GC
            nodeListFactory.unmake(nodeIterator);
        } else {
            throw new NodeError ("node can't be deleted: node not found in network");
        }
        
    }
    
    /**
     * Adds the given arc to this network between given two nodes. The arc should not belong to any other network and
     * the head and the tail node of the arc should both belong to the same network and be non-null otherwise 
     * <code>ArcError</code> will be thrown.
     *
     * @param addedArc  the arc to be added to this network
     * @param fromNode    the tail node to which the <code>addedArc</code> will be added
     * @param toNode    the head node to which the <code>addedArc</code> will be added
     */
     
    public void addArc(Arc addedArc,Node fromNode, Node toNode){
        
        ArcAlgInfo arcInfo;
        int i;
        
        if ( (addedArc.getTailNode()==null) && (addedArc.getHeadNode()==null) ) {
            if ( (fromNode.getContainerNetwork() == toNode.getContainerNetwork()) && (fromNode.getContainerNetwork()==this) ) {
                addedArc.setTailNode(fromNode);
                addedArc.setHeadNode(toNode);
                fromNode.addOutBoundArc(addedArc);
                toNode.addInBoundArc(addedArc);
                noOfArcs++;
        
                Vector arcAlgInfo = addedArc.getArcAlgorithmInfo();
                arcAlgInfo.setSize(algorithmList.size());
                for (i=algorithmList.size(); i-- > 0;) {
                    if (algorithmList.at(i)!=null) {
                        arcInfo=((Algorithm)(algorithmList.at(i))).getInstanceOfArcAlgInfo();
                        arcAlgInfo.setElementAt(arcInfo, i);
                    }        
                }
            } 
            else {
                throw new ArcError("arc can't be added: destination node/source node conflict with network");
            }
        }
        else {
            throw new ArcError("arc can't be added: arc already belongs to a structure");
        }
    }
        
    
    
    /**
     * Removes the given arc from this network. Failure to find the arc in this network results in <code>
     * ArcError</code> to be thrown.
     *
     * @param deletedArc    the arc to be deleted from this network
     * @param arcFactory    the ArcFactory to which the arc will be returned (can be null)
     */
    public void removeArc(Arc deletedArc, ArcFactory arcFactory){
        removeArc(deletedArc, arcFactory, true);
    }
    
    /**
     * Removes the given arc from this network. Failure to find the arc in this network results in <code>
     * ArcError</code> to be thrown.  <b>Warning:</b> Do not pass <code>true</code> for <code>collectArcInfo</code>
     * unless you really know what you're doing.  If this arc has been or will be cleaned, as in
     * <code>deletedArc.getTailNode().cleanOutBoundarcs(arcFactory)</code> or
     * <code>Network.clean</code>, it will be double-collected which will lead to bugs.  If you are unsure,
     * please use the corresponding public method without this parameter.
     *
     * @param deletedArc        the arc to be deleted from this network
     * @param arcFactory        the ArcFactory to which the arc will be returned (can be null)
     * @param collectArcInfo    if true, ArcAlgInfo will be collected to a factory
     */
    protected void removeArc(Arc deletedArc, ArcFactory arcFactory, boolean collectArcInfo){
        
        NodeList nodeIterator=nodeList.find(deletedArc.getTailNode());
        if (nodeIterator!=null) {
            
            nodeIterator=nodeList.find(deletedArc.getHeadNode());
            if (nodeIterator!=null) {
                Node tailNode;
                Node headNode;
                tailNode=deletedArc.getTailNode();
                headNode=deletedArc.getHeadNode();
                tailNode.removeOutBoundArc(deletedArc);
                headNode.removeInBoundArc(deletedArc);
                deletedArc.setHeadNode(null);
                deletedArc.setTailNode(null);
                if (collectArcInfo)
                    deletedArc.cleanArcAlgorithmInfo();
                if (arcFactory != null)
                    arcFactory.collectArc(deletedArc);
                noOfArcs--;
            }
        }
        else {
            Error err =  new ArcError("arc cadsfn't be deleted: arc does not belong to network");
            err.printStackTrace();
            throw err;
        }
            
    }
        
    
    
    
    
    /**
     * Redirects an arc in this network by changing its tail node. Failure to find arc, conflict between arc's
     * new/old head/tail nodes results in <code>ArcError</code> to be thrown.
     * @param   whichArc    arc whose tailnode is going to be changed
     * @param   newTailNode new tail node gor <code>whichArc</code>
     */
     
    public void changeArcTailNode(Arc whichArc, Node newTailNode) {
 
        Node myTail=whichArc.getTailNode();
        Node myHead=whichArc.getHeadNode();
        
        if ( (myTail!=null) && (myHead!=null) ) {
            
            if ( (myTail.getContainerNetwork() != null) && (myTail.getContainerNetwork() == newTailNode.getContainerNetwork()) )  {
                myTail.removeOutBoundArc(whichArc);
                newTailNode.addOutBoundArc(whichArc);
                whichArc.setTailNode(newTailNode);
            } 
            else {
                throw new ArcError("arc can't be redirected: new tail node's network conflicts with the old tail node network");
            }
        }
        else {
            throw new ArcError("arc can't be redirected:arc does not have tail/head node");
        }   
    }
    
    /**
     * Redirects an arc in this network by changing its head node. Failure to find arc, conflict between arc's
     * new/old head/tail nodes results in <code>ArcError</code> to be thrown.
     * @param   whichArc    arc whose tailnode is going to be changed
     * @param   newHeadNode new head node gor <code>whichArc</code>
     */
    public void changeArcHeadNode(Arc whichArc, Node newHeadNode) {

        Node myTail=whichArc.getTailNode();
        Node myHead=whichArc.getHeadNode();
        
        if ( (myTail!=null) && (myHead!=null) ) {
            
            if ( ( myHead.getContainerNetwork() == newHeadNode.getContainerNetwork() ) && (myHead.getContainerNetwork() != null) )  {
                myHead.removeInBoundArc(whichArc);
                newHeadNode.addInBoundArc(whichArc);
                whichArc.setHeadNode(newHeadNode);
            } 
            else {
                throw new ArcError("arc can't be redirected: new head node's network conflicts with the old node's nodes network");
            }
        }
        else {
            throw new ArcError("arc can't be redirected:arc does not have tail/head node");
        }
        
    }
       
   
    /** 
     * This method is used to simulate a circular linked list from the linear linked list provided by <code>
     * COM.objectspace.jgl</code> and should normally NOT be used by users.
     */
     
    public final NodeList getNextNodeInList(NodeList currentPosition){
        if (currentPosition != null) {
            return (currentPosition.next != null ? currentPosition.next : getNodeList());    
        }
        else
            return (getNodeList());
    }
    
    
    /** 
     * This method is used to simulate a circular linked list from the linear linked list provided by <code>
     * COM.objectspace.jgl</code> and should normally NOT be used by users.
     */
    public final NodeList getPreviousNodeInList(NodeList currentPosition){
        if (currentPosition != null) {
            return (currentPosition.prev != null ? currentPosition.prev : getNodeListEnd());    
        }
        else
            return (getNodeListEnd());
    }
    
    /**
     * Return the string representation of this object with number of nodes, number of arcs and number of algorithms all
     * seperated by spaces.
     */
     
    public String toString() {
        return Integer.toString(noOfNodes)+Integer.toString(noOfArcs)+Integer.toString(noOfAlgorithms);
    }        

}

